import json
import logging
import os
import traceback
from abc import ABC
from enum import Enum
from typing import List, Optional

import requests


Role = Enum("Role", ["system", "assistant", "user"])
Kind = Enum("Kind", ["chat", "text"])


class Log10Config:
    def __init__(
        self,
        url: str = None,
        token: str = None,
        org_id: str = None,
        tags: List[str] = None,
        DEBUG: bool = False,
    ):
        self.url = url if url else os.getenv("LOG10_URL")
        self.token = token if token else os.getenv("LOG10_TOKEN")
        self.org_id = org_id if org_id else os.getenv("LOG10_ORG_ID")
        self.DEBUG = DEBUG

        # Get tags from env, if not set, use empty list
        if tags:
            self.tags = tags
        elif os.getenv("LOG10_TAGS") is not None:
            self.tags = os.getenv("LOG10_TAGS").split(",")
        else:
            self.tags = []


class Message(ABC):
    def __init__(self, role: Role, content: str, id: str = None, completion: str = None):
        self.id = id
        self.role = role
        self.content = content
        self.completion = completion

    def to_dict(self):
        return {
            "role": self.role,
            "content": self.content,
        }

    def from_dict(message: dict):
        return Message(
            role=message["role"],
            content=message["content"],
            id=message.get("id"),
            completion=message.get("completion"),
        )


class Messages(ABC):
    def from_dict(messages: dict):
        return [Message.from_dict(message) for message in messages]


class Completion(ABC):
    pass


class ChatCompletion(Completion):
    def __init__(self, role: str, content: str, response: dict = None, completion_id: str = None):
        self.role = role
        self.content = content
        self.response = response
        self.completion_id = completion_id

    def to_dict(self) -> dict:
        return {
            "role": self.role,
            "content": self.content,
        }

    def __str__(self) -> str:
        return json.dumps(self.to_dict())


class TextCompletion(Completion):
    def __init__(self, text: str, response: dict = None, completion_id=None):
        self._text = text
        self.response = response
        self.completion_id = completion_id

    def text(self) -> str:
        return self._text

    def __str__(self) -> str:
        return self._text


class LLM(ABC):
    last_completion_response = None
    duration = None

    def __init__(self, hparams: dict = None, log10_config: Log10Config = None):
        self.log10_config = log10_config
        self.hparams = hparams

        # Start session
        if self.log10_config:
            session_url = self.log10_config.url + "/api/sessions"
            try:
                res = requests.request(
                    "POST",
                    session_url,
                    headers={
                        "x-log10-token": self.log10_config.token,
                        "Content-Type": "application/json",
                    },
                    json={"organization_id": self.log10_config.org_id},
                )
                response = res.json()
                self.session_id = response["sessionID"]
            except Exception as e:
                logging.warning(
                    f"Failed to start session with {session_url} using token {self.log10_config.token}. Won't be able to log. {e}"
                    + "\nSee https://github.com/log10-io/log10#%EF%B8%8F-setup for details."
                )
                self.log10_config = None
        else:
            logging.warning(
                "log10_config is not set. Won't be able to log with log10.io."
                + "\nSee https://github.com/log10-io/log10#%EF%B8%8F-setup for details."
            )

    def last_completion_url(self):
        if self.last_completion_response is None:
            return None

        return (
            self.log10_config.url
            + "/app/"
            + self.last_completion_response["organizationSlug"]
            + "/completions/"
            + self.last_completion_response["completionID"]
        )

    def last_duration(self):
        if self.duration is None:
            return None

        return self.duration

    def text(self, prompt: str, hparams: dict = None) -> TextCompletion:
        raise Exception("Not implemented")

    def text_request(self, prompt: str, hparams: dict = None) -> dict:
        raise Exception("Not implemented")

    def chat(self, messages: List[Message], hparams: dict = None) -> ChatCompletion:
        raise Exception("Not implemented")

    def chat_request(self, messages: List[Message], hparams: dict = None) -> dict:
        raise Exception("Not implemented")

    def api_request(self, rel_url: str, method: str, request: dict):
        return requests.request(
            method,
            f"{self.log10_config.url}{rel_url}",
            headers={
                "x-log10-token": self.log10_config.token,
                "Content-Type": "application/json",
            },
            json=request,
        )

    # Save the start of a completion in **openai request format**.
    def log_start(self, request, kind: Kind, tags: Optional[List[str]] = None):
        if not self.log10_config:
            return None

        res = self.api_request("/api/completions", "POST", {"organization_id": self.log10_config.org_id})
        self.last_completion_response = res.json()
        completion_id = res.json()["completionID"]

        # merge tags
        if tags:
            tags = list(set(tags + self.log10_config.tags))
        else:
            tags = self.log10_config.tags

        res = self.api_request(
            f"/api/completions/{completion_id}",
            "POST",
            {
                "kind": kind == Kind.text and "completion" or "chat",
                "organization_id": self.log10_config.org_id,
                "session_id": self.session_id,
                "orig_module": "openai.api_resources.completion"
                if kind == Kind.text
                else "openai.api_resources.chat_completion",
                "orig_qualname": "Completion.create" if kind == Kind.text else "ChatCompletion.create",
                "status": "started",
                "tags": tags,
                "request": json.dumps(request),
            },
        )

        return completion_id

    # Save the end of a completion in **openai request format**.
    def log_end(self, completion_id: str, response: dict, duration: int):
        if not self.log10_config:
            return None

        current_stack_frame = traceback.extract_stack()
        stacktrace = [
            {
                "file": frame.filename,
                "line": frame.line,
                "lineno": frame.lineno,
                "name": frame.name,
            }
            for frame in current_stack_frame
        ]

        duration_sec = int(duration * 1000)
        self.duration = duration_sec

        self.api_request(
            f"/api/completions/{completion_id}",
            "POST",
            {
                "organization_id": self.log10_config.org_id,
                "response": json.dumps(response),
                "status": "finished",
                "duration": duration_sec,
                "stacktrace": json.dumps(stacktrace),
            },
        )


class NoopLLM(LLM):
    def __init__(self, hparams: dict = None, log10_config=None):
        pass

    def chat(self, messages: List[Message], hparams: dict = None) -> ChatCompletion:
        logging.info("Received chat completion requst: " + str(messages))
        return ChatCompletion(role="assistant", content="I'm not a real LLM")

    def text(self, prompt: str, hparams: dict = None) -> TextCompletion:
        logging.info("Received text completion requst: " + prompt)
        return TextCompletion(text="I'm not a real LLM")



import asyncio
import json
import logging
import time
import traceback
import uuid
from datetime import datetime, timezone

import httpx
from httpx import Request, Response

from log10.llm import Log10Config
from log10.load import get_log10_session_tags, last_completion_response_var, session_id_var


logger: logging.Logger = logging.getLogger("LOG10")

GRAPHQL_URL = "https://graphql.log10.io/graphql"

_log10_config = Log10Config()
base_url = _log10_config.url
httpx_client = httpx.Client()
httpx_async_client = httpx.AsyncClient()


def _get_time_diff(created_at):
    time = datetime.fromisoformat(created_at)
    now = datetime.now(timezone.utc)
    diff = now - time
    # convert the time difference to human readable format
    if diff.days > 0:
        return f"{diff.days} days ago"
    elif diff.seconds > 3600:
        return f"{diff.seconds//3600} hours ago"
    elif diff.seconds > 60:
        return f"{diff.seconds//60} minutes ago"


def _try_get(url: str, timeout: int = 10) -> httpx.Response:
    httpx_client.headers = {
        "x-log10-token": _log10_config.token,
        "x-log10-organization-id": _log10_config.org_id,
        "Content-Type": "application/json",
    }
    httpx_timeout = httpx.Timeout(timeout)
    try:
        res = httpx_client.get(url, timeout=httpx_timeout)
        res.raise_for_status()
        return res
    except httpx.HTTPError as http_err:
        if "401" in str(http_err):
            logger.error(
                "Failed anthorization. Please verify that LOG10_TOKEN and LOG10_ORG_ID are set correctly and try again."
                + "\nSee https://github.com/log10-io/log10#%EF%B8%8F-setup for details"
            )
        else:
            logger.error(f"Failed with error: {http_err}")
        raise
    except Exception as e:
        logger.error(f"Error: {e}")
        if hasattr(e, "response") and hasattr(e.response, "json") and "error" in e.response.json():
            logger.error(e.response.json()["error"])
        raise


def _try_post_request(url: str, payload: dict = {}) -> httpx.Response:
    headers = {
        "x-log10-token": _log10_config.token,
        "x-log10-organization-id": _log10_config.org_id,
        "Content-Type": "application/json",
    }
    payload["organization_id"] = _log10_config.org_id
    res = None
    try:
        res = httpx_client.post(url, headers=headers, json=payload)
        res.raise_for_status()
        return res
    except httpx.HTTPError as http_err:
        if "401" in str(http_err):
            logger.error(
                "Failed anthorization. Please verify that LOG10_TOKEN and LOG10_ORG_ID are set correctly and try again."
                + "\nSee https://github.com/log10-io/log10#%EF%B8%8F-setup for details"
            )
        else:
            logger.error(f"Failed with error: {http_err}")
    except Exception as err:
        logger.error(f"Failed to insert in log10: {payload} with error {err}")


def _try_post_graphql_request(query: str, variables: dict = {}) -> httpx.Response:
    headers = {"content-type": "application/json", "x-api-token": _log10_config.token}

    payload = {"query": query, "variables": variables}
    res = None
    try:
        res = httpx_client.post(GRAPHQL_URL, headers=headers, json=payload)
        res.raise_for_status()
        return res
    except httpx.HTTPError as http_err:
        if "401" in str(http_err):
            logger.error(
                "Failed anthorization. Please verify that LOG10_TOKEN and LOG10_ORG_ID are set correctly and try again."
                + "\nSee https://github.com/log10-io/log10#%EF%B8%8F-setup for details"
            )
        else:
            logger.error(f"Failed with error: {http_err}")
    except Exception as err:
        logger.error(f"Failed to make requests to log10 graphQL: {payload} with error {err}")


async def _try_post_graphql_request_async(query: str, variables: dict = {}) -> httpx.Response:
    headers = {"content-type": "application/json", "x-api-token": _log10_config.token}

    payload = {"query": query, "variables": variables}
    res = None
    try:
        res = await httpx_async_client.post(GRAPHQL_URL, headers=headers, json=payload)
        res.raise_for_status()
        return res
    except httpx.HTTPError as http_err:
        if "401" in str(http_err):
            logger.error(
                "Failed anthorization. Please verify that LOG10_TOKEN and LOG10_ORG_ID are set correctly and try again."
                + "\nSee https://github.com/log10-io/log10#%EF%B8%8F-setup for details"
            )
        else:
            logger.error(f"Failed with error: {http_err}")
    except Exception as err:
        logger.error(f"Failed to make requests to log10 graphQL: {payload} with error {err}")


async def _try_post_request_async(url: str, payload: dict = {}) -> httpx.Response:
    headers = {
        "x-log10-token": _log10_config.token,
        "x-log10-organization-id": _log10_config.org_id,
        "Content-Type": "application/json",
    }
    payload["organization_id"] = _log10_config.org_id
    try:
        res = await httpx_async_client.post(url, headers=headers, json=payload)
        res.raise_for_status()
        return res
    except httpx.HTTPStatusError as http_err:
        if "401" in str(http_err):
            logger.error(
                "Failed authorization. Please verify that LOG10_TOKEN and LOG10_ORG_ID are set correctly and try again."
                + "\nSee https://github.com/log10-io/log10#%EF%B8%8F-setup for details"
            )
        else:
            logger.error(f"Failed with error: {http_err}")
    except Exception as err:
        logger.error(f"Failed to insert in log10: {payload} with error {err}")


def format_anthropic_tools_request(request_content) -> str:
    new_tools = []
    for tool in request_content["tools"]:
        new_tool = {
            "type": "function",
            "function": {"name": tool["name"], "description": tool["description"], "parameters": tool["input_schema"]},
        }
        new_tools.append(new_tool)
    request_content["tools"] = new_tools
    return json.dumps(request_content)


async def get_completion_id(request: Request):
    host = request.headers.get("host")
    if "anthropic" in host and "/v1/messages" not in str(request.url):
        logger.warning("Currently logging is only available for anthropic v1/messages.")
        return

    if "openai" in host and "v1/chat/completions" not in str(request.url):
        logger.warning("Currently logging is only available for openai v1/chat/completions.")
        return

    request.headers["x-log10-completion-id"] = str(uuid.uuid4())


async def log_request(request: Request):
    start_time = time.time()
    request.started = start_time
    completion_id = request.headers.get("x-log10-completion-id", "")
    if not completion_id:
        return

    last_completion_response_var.set({"completionID": completion_id})
    orig_module = ""
    orig_qualname = ""
    request_content_decode = request.content.decode("utf-8")
    host = request.headers.get("host")
    if "openai" in host:
        if "chat" in str(request.url):
            kind = "chat"
            orig_module = "openai.api_resources.chat_completion"
            orig_qualname = "ChatCompletion.create"
        else:
            kind = "completion"
            orig_module = "openai.api_resources.completion"
            orig_qualname = "Completion.create"
    elif "anthropic" in host:
        kind = "chat"
        request_content = json.loads(request_content_decode)
        if "tools" in request_content:
            orig_module = "anthropic.resources.beta.tools"
            orig_qualname = "Messages.stream"
            request_content_decode = format_anthropic_tools_request(request_content)
        else:
            orig_module = "anthropic.resources.messages"
            orig_qualname = "Messages.stream"
    else:
        logger.warning("Currently logging is only available for async openai and anthropic.")
        return
    log_row = {
        "status": "started",
        "kind": kind,
        "orig_module": orig_module,
        "orig_qualname": orig_qualname,
        "request": request_content_decode,
        "session_id": session_id_var.get(),
    }
    if get_log10_session_tags():
        log_row["tags"] = get_log10_session_tags()
    asyncio.create_task(_try_post_request_async(url=f"{base_url}/api/completions/{completion_id}", payload=log_row))


class _LogResponse(Response):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.full_content = ""
        self.function_name = ""
        self.full_argument = ""
        self.tool_calls = []
        self.finish_reason = ""

    async def aiter_bytes(self, *args, **kwargs):
        full_response = ""
        finished = False
        async for chunk in super().aiter_bytes(*args, **kwargs):
            full_response += chunk.decode(errors="ignore")
            if self.is_response_end_reached(full_response):
                finished = True
                duration = int(time.time() - self.request.started) * 1000

                completion_id = self.request.headers.get("x-log10-completion-id", "")
                if finished and completion_id:
                    current_stack_frame = traceback.extract_stack()
                    stacktrace = [
                        {
                            "file": frame.filename,
                            "line": frame.line,
                            "lineno": frame.lineno,
                            "name": frame.name,
                        }
                        for frame in current_stack_frame
                    ]

                    responses = full_response.split("\n\n")
                    r_json = self.parse_response_data(responses)

                    response_json = r_json.copy()
                    # r_json is the last response before "data: [DONE]"

                    # Choices can be empty list with the openai version > 1.26.0
                    if not response_json.get("choices"):
                        response_json["choices"] = [{"index": 0}]

                    if response_json.get("choices", []):
                        # It will only set finish_reason for openai version > 1.26.0
                        if self.finish_reason:
                            response_json["choices"][0]["finish_reason"] = self.finish_reason

                        if self.full_content:
                            response_json["choices"][0]["message"] = {
                                "role": "assistant",
                                "content": self.full_content,
                            }
                        elif self.tool_calls:
                            response_json["choices"][0]["message"] = {
                                "content": None,
                                "role": "assistant",
                                "tool_calls": self.tool_calls,
                            }
                        elif self.function_name and self.full_argument:
                            # function is deprecated in openai api
                            response_json["choices"][0]["function_call"] = {
                                "name": self.function_name,
                                "arguments": self.full_argument,
                            }

                    request_content_decode = self.request.content.decode("utf-8")
                    if "anthropic" in self.request.headers.get("host"):
                        request_content = json.loads(request_content_decode)
                        if "tools" in request_content:
                            request_content_decode = format_anthropic_tools_request(request_content)

                    log_row = {
                        "response": json.dumps(response_json),
                        "status": "finished",
                        "duration": duration,
                        "stacktrace": json.dumps(stacktrace),
                        "kind": "chat",
                        "request": request_content_decode,
                        "session_id": session_id_var.get(),
                    }
                    if get_log10_session_tags():
                        log_row["tags"] = get_log10_session_tags()
                    asyncio.create_task(
                        _try_post_request_async(url=f"{base_url}/api/completions/{completion_id}", payload=log_row)
                    )

            yield chunk

    def is_response_end_reached(self, text: str):
        host = self.request.headers.get("host")
        if "anthropic" in host:
            return self.is_anthropic_response_end_reached(text)
        elif "openai" in host:
            return self.is_openai_response_end_reached(text)
        else:
            logger.warning("Currently logging is only available for async openai and anthropic.")
            return False

    def is_anthropic_response_end_reached(self, text: str):
        return "event: message_stop" in text

    def is_openai_response_end_reached(self, text: str):
        return "data: [DONE]" in text

    def parse_anthropic_responses(self, responses: list[str]):
        message_id = None
        model = None
        finish_reason = None
        input_tokens = 0
        output_tokens = 0
        tool_call = {}
        arguments = ""
        for r in responses:
            if not r:
                break

            data_index = r.find("data:")
            r_json = json.loads(r[data_index + len("data:") :])

            ### anthropic first data contains
            ### {"event":"message_start","data":{"message":{"role":"user","content":"Hello, how are you today?"}}}
            type = r_json["type"]
            if type == "message_start":
                message_id = r_json["message"]["id"]
                model = r_json["message"]["model"]
                input_tokens = r_json["message"]["usage"]["input_tokens"]
            elif type == "content_block_start":
                content_block = r_json["content_block"]
                type = content_block["type"]
                if type == "tool_use":
                    id = content_block["id"]
                    tool_call = {
                        "id": id,
                        "type": "function",
                        "function": {"name": content_block["name"], "arguments": ""},
                    }

                if content_block_text := content_block.get("text", ""):
                    self.full_content += content_block_text
            elif type == "content_block_delta":
                delta = r_json["delta"]
                if delta_text := delta.get("text", ""):
                    self.full_content += delta_text

                if delta_partial_json := delta.get("partial_json", ""):
                    if self.full_content:
                        self.full_content += delta_partial_json
                    else:
                        arguments += delta_partial_json
            elif type == "message_delta":
                finish_reason = r_json["delta"]["stop_reason"]
                output_tokens = r_json["usage"]["output_tokens"]
            elif type == "content_block_end" or type == "message_end":
                if tool_call:
                    tool_call["function"]["arguments"] = arguments
                    self.tool_calls.append(tool_call)
                    tool_call = {}
                    arguments = ""

        return {
            "id": message_id,
            "object": "chat",
            "model": model,
            "choices": [
                {
                    "index": 0,
                    "finish_reason": finish_reason,
                }
            ],
            "usage": {
                "prompt_tokens": input_tokens,
                "completion_tokens": output_tokens,
                "total_tokens": input_tokens + output_tokens,
            },
        }

    def parse_openai_responses(self, responses: list[str]):
        r_json = {}
        for r in responses:
            if self.is_openai_response_end_reached(r):
                break

            # loading the substring of response text after 'data: '.
            # example: 'data: {"choices":[{"text":"Hello, how can I help you today?"}]}'
            data_index = r.find("data:")
            r_json = json.loads(r[data_index + len("data:") :])

            if r_json.get("choices", []):
                if delta := r_json["choices"][0].get("delta", {}):
                    # Delta may have content
                    # delta: { "content": " "}
                    if content := delta.get("content", ""):
                        self.full_content += content

                    # May be a function call, and have to reconstruct the arguments
                    if function_call := delta.get("function_call", {}):
                        # May be function name
                        if function_name := function_call.get("name", ""):
                            self.function_name = function_name
                        # May be function arguments
                        if arguments := function_call.get("arguments", ""):
                            self.full_argument += arguments

                    if tc := delta.get("tool_calls", []):
                        if tc[0].get("id", ""):
                            self.tool_calls.append(tc[0])
                        elif tc[0].get("function", {}).get("arguments", ""):
                            idx = tc[0].get("index")
                            self.tool_calls[idx]["function"]["arguments"] += tc[0]["function"]["arguments"]

                if fr := r_json["choices"][0].get("finish_reason", ""):
                    self.finish_reason = fr

        r_json["object"] = "chat.completion"

        return r_json

    def parse_response_data(self, responses: list[str]):
        host = self.request.headers.get("host")
        if "openai" in host:
            return self.parse_openai_responses(responses)
        elif "anthropic" in host:
            return self.parse_anthropic_responses(responses)
        else:
            logger.warning("Currently logging is only available for async openai and anthropic.")
            return None


class _LogTransport(httpx.AsyncBaseTransport):
    def __init__(self, transport: httpx.AsyncBaseTransport):
        self.transport = transport

    async def handle_async_request(self, request: httpx.Request) -> httpx.Response:
        try:
            response = await self.transport.handle_async_request(request)
        except Exception as e:
            logger.warning(f"Failed to send request: {e}")
            return

        if response.status_code >= 400:
            logger.warning(f"HTTP error occurred: {response.status_code}")
            return

        completion_id = request.headers.get("x-log10-completion-id", "")
        if not completion_id:
            return response

        if response.headers.get("content-type").startswith("application/json"):
            await response.aread()
            llm_response = response.json()

            current_stack_frame = traceback.extract_stack()
            stacktrace = [
                {
                    "file": frame.filename,
                    "line": frame.line,
                    "lineno": frame.lineno,
                    "name": frame.name,
                }
                for frame in current_stack_frame
            ]

            elapsed = time.time() - request.started
            if "anthropic" in request.url.host:
                from anthropic.types.beta.tools import (
                    ToolsBetaMessage,
                )

                from log10.anthropic import Anthropic

                llm_response = Anthropic.prepare_response(ToolsBetaMessage(**llm_response))

            log_row = {
                "response": json.dumps(llm_response),
                "status": "finished",
                "duration": int(elapsed * 1000),
                "stacktrace": json.dumps(stacktrace),
                "kind": "chat",
                "request": request.content.decode("utf-8"),
                "session_id": session_id_var.get(),
            }
            if get_log10_session_tags():
                log_row["tags"] = get_log10_session_tags()
            asyncio.create_task(
                _try_post_request_async(url=f"{base_url}/api/completions/{completion_id}", payload=log_row)
            )
            return response
        elif response.headers.get("content-type").startswith("text/event-stream"):
            return _LogResponse(
                status_code=response.status_code,
                headers=response.headers,
                stream=response.stream,
                extensions=response.extensions,
                request=request,
            )

        # In case of an error, get out of the way
        return response


async def finalize():
    pending = asyncio.all_tasks()
    pending.remove(asyncio.current_task())
    await asyncio.gather(*pending)



import json
import logging

import httpx
from dotenv import load_dotenv
from rich.console import Console
from rich.panel import Panel
from rich.text import Text

from log10.llm import Log10Config


load_dotenv()

logging.basicConfig(
    format="[%(asctime)s - %(name)s - %(levelname)s] %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
)
logger: logging.Logger = logging.getLogger("LOG10")
logger.setLevel(logging.INFO)


class PromptAnalyzer:
    convert_url = "/api/experimental/autoprompt/convert"
    report_url = "/api/experimental/autoprompt/report"
    suggestions_url = "/api/experimental/autoprompt/suggestions"

    def __init__(self, log10_config: Log10Config = None):
        self._prompt_history: list[str] = []
        self._converted_prompt_history: list[dict] = []
        self._suggestions_history: list[dict] = []
        self._report_history: list[dict] = []

        self._log10_config = log10_config or Log10Config()
        self._http_client = httpx.Client()

        # Set by `_convert`.
        self.__session_id = None

    def _post_request(self, url: str, json_payload: dict) -> httpx.Response:
        headers = {"x-log10-token": self._log10_config.token, "Content-Type": "application/json"}
        json_payload["organization_id"] = self._log10_config.org_id
        try:
            timeout = httpx.Timeout(timeout=5, connect=5, read=5 * 60, write=5)
            res = self._http_client.post(
                self._log10_config.url + url, headers=headers, json=json_payload, timeout=timeout
            )
            res.raise_for_status()
            return res
        except Exception as e:
            logger.error(e)
            raise

    def _convert(self, prompt: str) -> dict:
        json_payload = {"prompt": prompt}

        # Pass session ID if it has already been set.
        if self.__session_id:
            json_payload["session_id"] = self.__session_id

        res = self._post_request(self.convert_url, json_payload)
        res_json = res.json()
        converted = res_json.get("output")

        # The convert API returns a session ID we can use to link together a session.
        if not self.__session_id:
            self.__session_id = res_json.get("session_id")

        return converted

    def _report(self, last_prompt: dict, current_prompt: dict, suggestions: dict) -> dict:
        json_payload = {
            "base_prompt": json.dumps(last_prompt),
            "new_prompt": json.dumps(current_prompt),
            "suggestions": json.dumps(suggestions),
            "session_id": self.__session_id,
        }
        res = self._post_request(self.report_url, json_payload)
        report = res.json().get("output")
        return report

    def _suggest(self, prompt_json: json, report: dict | None = None) -> dict:
        if report is None or report == {}:
            report = "[{}]"
        else:
            report = json.dumps(report)

        json_payload = {
            "base_prompt": json.dumps(prompt_json),
            "report": report,
            "session_id": self.__session_id,
        }
        res = self._post_request(self.suggestions_url, json_payload)
        suggestion = res.json().get("output")
        return suggestion

    def analyze(self, prompt: str) -> dict:
        """
        Analyze prompt and return suggestions. You can make multiple calls to this method by
        making changes to the prompt. This function keeps track of the prompt history and generate
        suggestions based on the prompt history.

        Example:
        >>> from log10.prompt_analyzer import PromptAnalyzer
        >>> analyzer = PromptAnalyzer()
        >>> prompt = "You are an assistant communicating with XYZ users on their support line."
        >>> suggestions = analyzer.analyze(prompt)
        >>> print(suggestions)
        """
        total_steps = 3 if self._suggestions_history else 2
        step = 0
        try:
            step += 1
            logger.info(f"[Step {step}/{total_steps}] Analyzing prompt...")

            self._prompt_history.append(prompt)
            converted_prompt = self._convert(prompt)
            self._converted_prompt_history.append(converted_prompt)

            logger.debug(f"converted prompt: {converted_prompt}")

            if self._suggestions_history:
                assert self._converted_prompt_history, "Prompt history is empty."
                assert self._suggestions_history, "Suggestions history is empty."

                step += 1
                logger.info(f"[Step {step}/{total_steps}] Generating report...")

                last_converted_prompt = self._converted_prompt_history[-1]
                last_suggestions = self._suggestions_history[-1]
                report_dict = self._report(last_converted_prompt, converted_prompt, last_suggestions)
                self._report_history.append(report_dict)

                logger.debug(f"report: {report_dict}")

            step += 1
            logger.info(f"[Step {step}/{total_steps}] Generating suggestions...")

            last_report = self._report_history[-1] if self._report_history else {}
            suggestions_dict = self._suggest(converted_prompt, last_report)
            self._suggestions_history.append(suggestions_dict)

            logger.debug(f"suggestions: {suggestions_dict}")
            logger.info("Done analyzing.")

            return suggestions_dict
        except Exception as e:
            logger.error(e)


def convert_suggestion_to_markdown(suggestion):
    markdown_text = ""

    markdown_text += "| Category | Recommendation |\n"
    markdown_text += "| --- | --- |\n"
    for category, contents in suggestion.items():
        for subcategory, details in contents.items():
            color = details["color"]  # Assuming 'color' is always provided
            recommendation = details["recommendation"]

            # Using HTML to colorize the subcategory title if 'color' is specified
            colored_subcategory = (
                f'<span style="color:{color};">{subcategory}</span> <br> <span style="color:{color};"></span>'
            )

            # Add each subcategory and its recommendation to the table
            markdown_text += f"| **{colored_subcategory}** | {recommendation} |\n"

        # markdown_text += "\n"  # Add a newline for spacing before the next section

    return markdown_text


def display_prompt_analyzer_suggestions(data, title=""):
    console = Console()
    console.print(f"[bold magenta]{title}[/bold magenta]")
    for key, value in data.items():
        for sub_key, sub_value in value.items():
            text = Text(justify="left")
            text.append(f"Recommendation: {sub_value['recommendation']}", style="white")
            console.print(
                Panel(
                    text, title=f"[bold {sub_value['color']}]{sub_key}[/]", border_style=f"bold {sub_value['color']}"
                )
            )



import logging
import time
from typing import List

import anthropic
from anthropic import AI_PROMPT, HUMAN_PROMPT
from anthropic.types.beta.tools import (
    ToolsBetaMessage,
)

from log10.llm import LLM, ChatCompletion, Kind, Message, TextCompletion
from log10.utils import merge_hparams


class Anthropic(LLM):
    def __init__(self, hparams: dict = None, skip_initialization: bool = False, log10_config=None):
        super().__init__(hparams, log10_config)

        if not skip_initialization:
            self.client = anthropic.Anthropic()
        self.hparams = hparams

        if "max_tokens_to_sample" not in self.hparams:
            self.hparams["max_tokens_to_sample"] = 1024

    def chat(self, messages: List[Message], hparams: dict = None) -> ChatCompletion:
        """
        Example:
            >>> from log10.llm import Log10Config, Message
            >>> from log10.anthropic import Anthropic
            >>> llm = Anthropic({"model": "claude-1"}, log10_config=Log10Config())
            >>> response = llm.chat([Message(role="user", content="Hello, how are you?")])
            >>> print(response)
            >>> print(f"Duration: {llm.last_duration()}")
        """
        chat_request = self.chat_request(messages, hparams)

        openai_request = {
            "messages": [message.to_dict() for message in messages],
            **chat_request,
        }

        completion_id = self.log_start(openai_request, Kind.chat)

        start_time = time.perf_counter()
        completion = self.client.completions.create(**chat_request)
        content = completion.completion

        response = Anthropic.prepare_response(completion, chat_request["prompt"])

        self.log_end(
            completion_id,
            response,
            time.perf_counter() - start_time,
        )

        return ChatCompletion(role="assistant", content=content, response=response)

    def chat_request(self, messages: List[Message], hparams: dict = None) -> dict:
        merged_hparams = merge_hparams(hparams, self.hparams)

        # NOTE: That we may have to convert this to openai messages, if we want
        #       to use the same log viewer for all chat based models.
        prompt = Anthropic.convert_history_to_claude(messages)
        return {"prompt": prompt, "stop_sequences": [HUMAN_PROMPT], **merged_hparams}

    def text(self, prompt: str, hparams: dict = None) -> TextCompletion:
        """
        Example:
            >>> from log10.llm import Log10Config
            >>> from log10.anthropic import Anthropic
            >>> llm = Anthropic({"model": "claude-1"}, log10_config=Log10Config())
            >>> response = llm.text("Foobarbaz")
            >>> print(response)
            >>> print(f"Duration: {llm.last_duration()}")
        """
        text_request = self.text_request(prompt, hparams)

        openai_request = {
            **text_request,
            "prompt": prompt,
        }

        start_time = time.perf_counter()
        completion_id = self.log_start(openai_request, Kind.text)
        completion = self.client.completions.create(**text_request)
        text = completion.completion

        response = Anthropic.prepare_response(completion, text_request["prompt"])

        self.log_end(
            completion_id,
            response,
            time.perf_counter() - start_time,
        )

        logging.info("Returning text completion")
        return TextCompletion(text=text, response=response)

    def text_request(self, prompt: str, hparams: dict = None) -> TextCompletion:
        merged_hparams = merge_hparams(hparams, self.hparams)
        return {
            "prompt": HUMAN_PROMPT + prompt + "\n" + AI_PROMPT,
            "stop_sequences": [HUMAN_PROMPT],
            **merged_hparams,
        }

    def convert_history_to_claude(messages: List[Message]):
        text = ""
        for message in messages:
            # Anthropic doesn't support a system prompt OOB
            if message.role == "user" or message.role == "system":
                text += HUMAN_PROMPT
            elif message.role == "assistant":
                text += AI_PROMPT
            text += f"{message.content}"
        text += AI_PROMPT
        return text

    def convert_claude_to_messages(prompt: str):
        pass

    @staticmethod
    def create_tokens_usage(prompt: str, completion: str):
        client = anthropic.Anthropic()
        prompt_tokens = client.count_tokens(prompt)
        completion_tokens = client.count_tokens(completion)
        total_tokens = prompt_tokens + completion_tokens

        # Imitate OpenAI usage format.
        return {
            "prompt_tokens": prompt_tokens,
            "completion_tokens": completion_tokens,
            "total_tokens": total_tokens,
        }

    @staticmethod
    def prepare_response(
        response: anthropic.types.Completion | anthropic.types.Message | ToolsBetaMessage, input_prompt: str = ""
    ) -> dict:
        if not hasattr(response, "stop_reason"):
            return None

        if response.stop_reason in ["stop_sequence", "end_turn", "tool_use"]:
            reason = "stop"
        elif response.stop_reason == "max_tokens":
            reason = "length"

        ret_response = {
            "id": response.id,
            "object": "completion",
            "model": response.model,
            "choices": [
                {
                    "index": 0,
                    "finish_reason": reason,
                }
            ],
        }

        if isinstance(response, anthropic.types.Message):
            tokens_usage = {
                "prompt_tokens": response.usage.input_tokens,
                "completion_tokens": response.usage.output_tokens,
                "total_tokens": response.usage.input_tokens + response.usage.output_tokens,
            }
            ret_response["choices"][0]["message"] = {"role": response.role, "content": response.content[0].text}
        elif isinstance(response, anthropic.types.Completion):
            tokens_usage = Anthropic.create_tokens_usage(input_prompt, response.completion)
            ret_response["choices"][0]["text"] = response.completion
        elif isinstance(response, ToolsBetaMessage):
            tokens_usage = {
                "prompt_tokens": response.usage.input_tokens,
                "completion_tokens": response.usage.output_tokens,
                "total_tokens": response.usage.input_tokens + response.usage.output_tokens,
            }
            ret_response["choices"][0]["message"] = {"role": response.role, "content": response.content[0].text}
        ret_response["usage"] = tokens_usage

        return ret_response



from pandas import DataFrame
from tabulate import tabulate


def generate_results_table(dataframe: DataFrame, column_list: list[str] = None, section_name: str = "") -> str:
    selected_df = dataframe[column_list] if column_list else dataframe
    section_name = f"## {section_name}" if section_name else "## Test Results"

    table = tabulate(selected_df, headers="keys", tablefmt="pipe", showindex=True)
    ret_str = f"{section_name}\n{table}"
    return ret_str


def generate_markdown_report(test_name: str, report_strings: list[str]):
    with open(test_name, "w") as f:
        f.write(f"Generated from {test_name}.\n\n")
        for report_string in report_strings:
            f.write(report_string + "\n\n")



from log10.llm import LLM, Message
from log10.tools import browser


# Set up Summarizer agent
system_prompt = "You are an expert at extracting the main points from a website. Only look at the provided website content by the user to extract the main points."
summarize_prompt = "Extract the main points from the following website:\n {website_text}"


def scrape_summarizer(url, llm: LLM):
    website_text = browser(url)
    prompt = summarize_prompt.format(website_text=website_text)
    messages = [
        Message(role="system", content=system_prompt),
        Message(role="user", content=prompt),
    ]

    hparams = {"temperature": 0.2}

    completion = llm.chat(messages, hparams)

    return completion.content



import json
import logging

import click
import httpx
from dotenv import load_dotenv
from rich.console import Console
from rich.table import Table

from log10._httpx_utils import _get_time_diff, _try_get
from log10.llm import Log10Config


load_dotenv()

logging.basicConfig(
    format="[%(asctime)s - %(name)s - %(levelname)s] %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
)
logger: logging.Logger = logging.getLogger("LOG10")
logger.setLevel(logging.INFO)


class FeedbackTask:
    feedback_task_create_url = "/api/v1/feedback_task"

    def __init__(self, log10_config: Log10Config = None):
        self._log10_config = log10_config or Log10Config()
        self._http_client = httpx.Client()
        self._http_client.headers = {
            "x-log10-token": self._log10_config.token,
            "x-log10-organization-id": self._log10_config.org_id,
            "Content-Type": "application/json",
        }

    def _post_request(self, url: str, json_payload: dict) -> httpx.Response:
        json_payload["organization_id"] = self._log10_config.org_id
        try:
            res = self._http_client.post(self._log10_config.url + url, json=json_payload)
            res.raise_for_status()
            return res
        except Exception as e:
            logger.error(e)
            logger.error(e.response.json()["error"])
            raise

    def create(
        self, task_schema: dict, name: str, completion_tags_selector: list[str] = None, instruction: str = None
    ) -> httpx.Response:
        json_payload = {"json_schema": task_schema, "name": name, "completion_tags_selector": completion_tags_selector}
        if instruction:
            json_payload["instruction"] = instruction

        res = self._post_request(self.feedback_task_create_url, json_payload)
        return res

    def list(self, limit: int = 10, offset: int = 0) -> httpx.Response:
        url = f"{self._log10_config.url}/api/v1/feedback_task?organization_id={self._log10_config.org_id}&offset={offset}&limit={limit}"
        try:
            res = self._http_client.get(url=url)
            res.raise_for_status()
            return res
        except Exception as e:
            logger.error(e)
            logger.error(e.response.json()["error"])
            raise

    def get(self, id: str) -> httpx.Response:
        based_url = self._log10_config.url
        api_url = "/api/v1/feedback_task"
        get_url = f"{based_url}{api_url}/{id}?organization_id={self._log10_config.org_id}"
        res = _try_get(get_url)
        res.raise_for_status()
        if res.status_code != 200:
            raise Exception(f"Error fetching feedback task {res.json()}")
        return res


# create a cli interface for FeebackTask.create function
@click.command()
@click.option("--name", prompt="Enter feedback task name", help="Name of the task")
@click.option("--task_schema", prompt="Enter feedback task schema", help="Task schema")
@click.option("--instruction", help="Task instruction", default="")
@click.option(
    "--completion_tags_selector",
    help="Completion tags selector",
)
def create_feedback_task(name, task_schema, instruction, completion_tags_selector=None):
    click.echo("Creating feedback task")
    tags = []

    if completion_tags_selector:
        tags = completion_tags_selector.split(",")

    task_schema = json.loads(task_schema)
    task = FeedbackTask().create(
        name=name, task_schema=task_schema, completion_tags_selector=tags, instruction=instruction
    )
    click.echo(f"Use this task_id to add feedback: {task.json()['id']}")


@click.command()
@click.option("--limit", default=25, help="Number of feedback tasks to fetch")
@click.option("--offset", default=0, help="Offset for the feedback tasks")
def list_feedback_task(limit, offset):
    res = FeedbackTask().list(limit=limit, offset=offset)
    feedback_tasks = res.json()

    data_for_table = []

    for task in feedback_tasks["data"]:
        data_for_table.append(
            {
                "id": task["id"],
                "created_at": _get_time_diff(task["created_at"]),
                "name": task["name"],
                "required": task["json_schema"]["required"],
                "instruction": task["instruction"],
            }
        )

    table = Table(title="Feedback Tasks")
    table.add_column("ID", style="dim")
    table.add_column("Created At")
    table.add_column("Name")
    table.add_column("Required")
    table.add_column("Instruction")
    for item in data_for_table:
        required = ", ".join(item["required"]) if item["required"] else ""
        table.add_row(item["id"], item["created_at"], item["name"], required, item["instruction"])

    console = Console()
    console.print(table)


@click.command()
@click.option("--id", help="Get feedback task by ID")
def get_feedback_task(id):
    try:
        res = FeedbackTask().get(id)
    except Exception as e:
        click.echo(f"Error fetching feedback task {e}")
        if hasattr(e, "response") and hasattr(e.response, "json") and "error" in e.response.json():
            click.echo(e.response.json()["error"])
        return
    task = json.dumps(res.json())
    console = Console()
    console.print_json(task)



import time
import uuid

from mcli import predict

from log10.llm import LLM, Kind, Log10Config, TextCompletion
from log10.utils import merge_hparams


def llama_2_70b_chat(prompt: str, hparams: dict = None, log10_config: Log10Config = None) -> str:
    """
    Example:
        >>> from log10.llm import Log10Config
        >>> from log10.mosaicml import llama_2_70b_chat
        >>> response = llama_2_70b_chat("Hello, how are you?", {"temperature": 0.3, "max_new_tokens": 10}, log10_config=Log10Config())
        >>> print(response)
    """
    return MosaicML({"model": "llama2-70b-chat/v1"}, log10_config).text(prompt, hparams)


class MosaicML(LLM):
    """
    Example:
        >>> from log10.llm import Log10Config
        >>> from log10.mosaicml import MosaicML
        >>> llm = MosaicML({"model": "llama2-70b-chat/v1"}, log10_config=Log10Config())
        >>> response = llm.text("Hello, how are you?", {"temperature": 0.3, "max_new_tokens": 10})
        >>> print(response)

    For more information on MosaicML Complete, see https://docs.mosaicml.com/en/latest/inference.html#text-completion-request-endpoint
    """

    mosaicml_host_url = "https://models.hosted-on.mosaicml.hosting/"

    def __init__(self, hparams: dict = None, log10_config: Log10Config = None):
        super().__init__(hparams, log10_config)

    def text(self, prompt: str, hparams: dict = None) -> TextCompletion:
        request = self.text_request(prompt, hparams)
        self.model = request.get("model")
        openai_request = {
            **request,
            "model": f"mosaicml/{self.model}",
        }

        start_time = time.perf_counter()
        self.completion_id = self.log_start(openai_request, Kind.text)
        completion = predict(
            f"{self.mosaicml_host_url}{self.model}",
            {
                "inputs": [prompt],
                "parameters": request,
            },
        )

        response = self._prepare_response(completion)

        self.log_end(
            self.completion_id,
            response,
            time.perf_counter() - start_time,
        )

        return TextCompletion(text=response["choices"][0]["text"], response=response)

    def text_request(self, prompt: str, hparams: dict = None) -> dict:
        merged_hparams = merge_hparams(hparams, self.hparams)
        return {"prompt": prompt, **merged_hparams}

    def _prepare_response(self, completion: dict) -> dict:
        response = {
            "id": str(uuid.uuid4()),
            "object": "text_completion",
            "model": f"mosaicml/{self.model}",
            "choices": [
                {
                    "index": 0,
                    "text": completion["outputs"][0],
                    "logprobs": None,
                    "finish_reason": None,
                }
            ],
            "usage": None,
        }
        return response



import time
import uuid
from uuid import UUID

from langchain.schema import AIMessage, BaseMessage, HumanMessage, SystemMessage


"""Callback Handler that prints to std out."""
import logging
from typing import Any, Dict, List, Optional, Union

from langchain.callbacks.base import BaseCallbackHandler
from langchain.schema import AgentAction, AgentFinish, LLMResult

from log10.llm import LLM, Kind, Message


logging.basicConfig()
logger = logging.getLogger("log10")


def kwargs_to_hparams(kwargs: Dict[str, Any]) -> Dict[str, Any]:
    """Convert kwargs to hparams."""
    hparams = {}
    if "temperature" in kwargs:
        hparams["temperature"] = kwargs["temperature"]
    if "top_p" in kwargs:
        hparams["top_p"] = kwargs["top_p"]
    if "top_k" in kwargs:
        hparams["top_k"] = kwargs["top_k"]
    if "max_tokens" in kwargs:
        hparams["max_tokens"] = kwargs["max_tokens"]
    if "max_tokens_to_sample" in kwargs:
        hparams["max_tokens"] = kwargs["max_tokens_to_sample"]
    if "frequency_penalty" in kwargs:
        hparams["frequency_penalty"] = kwargs["frequency_penalty"]
    if "presence_penalty" in kwargs:
        hparams["presence_penalty"] = kwargs["presence_penalty"]
    return hparams


def get_log10_messages(langchain_messages: List[BaseMessage]) -> List[Message]:
    role_map = {AIMessage: "assistant", HumanMessage: "user", SystemMessage: "system"}

    for m in langchain_messages:
        logger.debug(f"message: {m}")
        if type(m) not in role_map:
            raise BaseException(f"Unsupported message type {type(m)}. Supported types: {list(role_map.values())}")

    return [Message(role=role_map[type(m)], content=m.content) for m in langchain_messages]


class Log10Callback(BaseCallbackHandler, LLM):
    """Callback Handler that prints to std out."""

    def __init__(self, log10_config: Optional[dict] = None) -> None:
        """Initialize callback handler."""
        super().__init__(log10_config=log10_config, hparams=None)
        self.runs = {}

        if log10_config.DEBUG:
            logger.setLevel(logging.DEBUG)
        else:
            logger.setLevel(logging.INFO)

    def on_llm_start(
        self,
        serialized: Dict[str, Any],
        prompts: List[str],
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        tags: Optional[List[str]] = None,
        **kwargs: Any,
    ) -> None:
        """Print out the prompts."""
        logger.debug(
            f"**\n**on_llm_start**\n**\n: serialized:\n {serialized} \n\n prompts:\n {prompts} \n\n rest: {kwargs}"
        )
        kwargs = serialized.get("kwargs", {})
        hparams = kwargs_to_hparams(kwargs)

        model = kwargs.get("model_name", None)
        if model is None:
            model = kwargs.get("model", None)
        if model is None:
            raise BaseException("No model found in serialized or kwargs")

        if len(prompts) != 1:
            raise BaseException("Only support one prompt at a time")

        request = {"model": model, "prompt": prompts[0], **hparams}

        logger.debug(f"request: {request}")

        completion_id = self.log_start(request, Kind.text, tags)

        self.runs[run_id] = {
            "kind": Kind.text,
            "completion_id": completion_id,
            "start_time": time.perf_counter(),
            "model": model,
        }

    def on_chat_model_start(
        self,
        serialized: Dict[str, Any],
        messages: List[List[BaseMessage]],
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        tags: Optional[List[str]] = None,
        **kwargs: Any,
    ) -> None:
        logger.debug(
            f"**\n**on_chat_model_start**\n**\n: run_id:{run_id}\nserialized:\n{serialized}\n\nmessages:\n{messages}\n\nkwargs: {kwargs}"
        )

        #
        # Find model string
        #
        kwargs = serialized.get("kwargs", {})
        model = kwargs.get("model_name", None)
        if model is None:
            model = kwargs.get("model", None)
        if model is None:
            raise BaseException("No model found in serialized or kwargs")

        hparams = kwargs_to_hparams(kwargs)
        hparams["model"] = model

        logger.debug(f"hparams: {hparams}")

        if len(messages) != 1:
            raise BaseException("Only support one message at a time")

        # Convert messages to log10 format
        log10_messages = get_log10_messages(messages[0])

        request = {
            "messages": [message.to_dict() for message in log10_messages],
            **hparams,
        }
        logger.debug(f"request: {request}")

        completion_id = self.log_start(
            request,
            Kind.chat,
            tags,
        )

        self.runs[run_id] = {
            "kind": Kind.chat,
            "completion_id": completion_id,
            "start_time": time.perf_counter(),
            "model": model,
        }

        logger.debug(f"logged start with completion_id: {completion_id}")

    def on_llm_end(
        self,
        response: LLMResult,
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        **kwargs: Any,
    ) -> None:
        """Do nothing."""
        # Find run in runs.
        run = self.runs.get(run_id, None)
        if run is None:
            raise BaseException("Could not find run in runs")

        if run["kind"] != Kind.chat and run["kind"] != Kind.text:
            raise BaseException("Only support chat kind")

        duration = time.perf_counter() - run["start_time"]

        # Log end
        if len(response.generations) != 1:
            raise BaseException("Only support one message at a time")
        if len(response.generations[0]) != 1:
            raise BaseException("Only support one message at a time")

        content = response.generations[0][0].text

        log10response = {}
        if run["kind"] == Kind.chat:
            log10response = {
                "id": str(uuid.uuid4()),
                "object": "chat.completion",
                "model": run["model"],
                "choices": [
                    {
                        "index": 0,
                        "message": {"role": "assistant", "content": content},
                        "finish_reason": "stop",
                    }
                ],
            }
        elif run["kind"] == Kind.text:
            log10response = {
                "id": str(uuid.uuid4()),
                "object": "text_completion",
                "model": run["model"],
                "choices": [
                    {
                        "index": 0,
                        "text": content,
                        "logprobs": None,
                        "finish_reason": "stop",
                    }
                ],
            }

        # Determine if we can provide usage metrics (token count).
        logger.debug(f"**** response: {response}")
        if response.llm_output is not None:
            token_usage = response.llm_output.get("token_usage")
            if token_usage is not None:
                log10response["usage"] = token_usage
                logger.debug(f"usage: {log10response['usage']}")

        logger.debug(f"**\n**on_llm_end**\n**\n: response:\n {log10response} \n\n rest: {kwargs}")
        self.log_end(run["completion_id"], log10response, duration)

    def on_llm_new_token(self, token: str, **kwargs: Any) -> None:
        """Do nothing."""
        logger.debug(f"token:\n {token} \n\n rest: {kwargs}")

    def on_llm_error(self, error: Union[Exception, KeyboardInterrupt], **kwargs: Any) -> None:
        """Do nothing."""
        logger.debug(f"error:\n {error} \n\n rest: {kwargs}")

    def on_chain_start(self, serialized: Dict[str, Any], inputs: Dict[str, Any], **kwargs: Any) -> None:
        """Print out that we are entering a chain."""
        pass

    def on_chain_end(self, outputs: Dict[str, Any], **kwargs: Any) -> None:
        """Print out that we finished a chain."""
        pass

    def on_chain_error(self, error: Union[Exception, KeyboardInterrupt], **kwargs: Any) -> None:
        """Do nothing."""
        pass

    def on_tool_start(
        self,
        serialized: Dict[str, Any],
        input_str: str,
        **kwargs: Any,
    ) -> None:
        """Do nothing."""
        pass

    def on_agent_action(self, action: AgentAction, color: Optional[str] = None, **kwargs: Any) -> Any:
        """Run on agent action."""
        pass

    def on_tool_end(
        self,
        output: str,
        color: Optional[str] = None,
        observation_prefix: Optional[str] = None,
        llm_prefix: Optional[str] = None,
        **kwargs: Any,
    ) -> None:
        """If not the final action, print out observation."""
        pass

    def on_tool_error(self, error: Union[Exception, KeyboardInterrupt], **kwargs: Any) -> None:
        """Do nothing."""
        pass

    def on_text(
        self,
        text: str,
        color: Optional[str] = None,
        end: str = "",
        **kwargs: Any,
    ) -> None:
        """Run when agent ends."""
        pass

    def on_agent_finish(self, finish: AgentFinish, color: Optional[str] = None, **kwargs: Any) -> None:
        """Run on agent end."""
        pass



import click

from log10.completions.completions import benchmark_models, download_completions, get_completion, list_completions
from log10.feedback.autofeedback import auto_feedback_icl, get_autofeedback_cli
from log10.feedback.feedback import create_feedback, download_feedback, get_feedback, list_feedback
from log10.feedback.feedback_task import create_feedback_task, get_feedback_task, list_feedback_task


@click.group()
def cli():
    pass


@click.group()
def completions():
    """
    Manage logs from completions i.e. logs from users
    """
    pass


@click.group(name="feedback")
def feedback():
    """
    Manage feedback for completions i.e. capturing feedback from users
    """
    pass


@click.group(name="auto_feedback")
def auto_feedback():
    """
    Manage auto feedback for completions i.e. capturing feedback from users
    """
    pass


@click.group()
def feedback_task():
    """
    Manage tasks for feedback i.e. instructions and schema for feedback
    """
    pass


cli.add_command(completions)
completions.add_command(list_completions, "list")
completions.add_command(get_completion, "get")
completions.add_command(download_completions, "download")
completions.add_command(benchmark_models, "benchmark_models")

cli.add_command(feedback)
feedback.add_command(create_feedback, "create")
feedback.add_command(list_feedback, "list")
feedback.add_command(get_feedback, "get")
feedback.add_command(download_feedback, "download")
feedback.add_command(auto_feedback_icl, "predict")
feedback.add_command(auto_feedback, "autofeedback")
# Subcommands for auto_feedback under feedback command
auto_feedback.add_command(get_autofeedback_cli, "get")

cli.add_command(feedback_task)
feedback_task.add_command(create_feedback_task, "create")
feedback_task.add_command(list_feedback_task, "list")
feedback_task.add_command(get_feedback_task, "get")

if __name__ == "__main__":
    cli()



import logging
from typing import List, Optional

from log10.llm import LLM, Kind, Log10Config
from log10.load import last_completion_response_var


try:
    from litellm.integrations.custom_logger import CustomLogger
except ImportError as error:
    msg = "To use Log10LitellmLogger to log LiteLLM completions, please install litellm using 'pip install log10-io[litellm]'"
    raise ImportError(msg) from error

logging.basicConfig()
logger = logging.getLogger("log10")


class Log10LitellmLogger(CustomLogger, LLM):
    """
    Example:
        >>> import litellm
        >>> from log10.litellm import Log10LitellmLogger
        >>> log10_handler = Log10LitellmLogger()
        >>> litellm.callbacks = [log10_handler]
        >>> response = litellm.completion(model="gpt-3.5-turbo", messages=[{ "role": "user", "content": "Count to 10."}], stream=True)
        >>> for chunk in response:
        >>>     if chunk.choices[0].delta.content:
        >>>          print(chunk.choices[0].delta.content, end="", flush=True)
    Example:
        >>> import litellm
        >>> from log10.litellm import Log10LitellmLogger
        >>> log10_handler = Log10LitellmLogger()
        >>> litellm.callbacks = [log10_handler]
        >>> response = litellm.completion(model="gpt-3.5-turbo", messages=[{ "role": "user", "content": "What's the next number, 1, 3, 5, 7? Say the number only."}])
        >>> print(response.choices[0].message.content)
    """

    runs = {}
    tags = []

    def __init__(self, log10_config: Optional[dict] = None, tags: List[str] = []) -> None:
        CustomLogger.__init__(self)

        log10_config = log10_config or Log10Config()
        self.tags = tags
        LLM.__init__(self, log10_config=log10_config, hparams=None)
        if log10_config.DEBUG:
            logger.setLevel(logging.DEBUG)
        else:
            logger.setLevel(logging.INFO)

    def log_pre_api_call(self, model, messages, kwargs):
        logger.debug(
            f"**\n**log_pre_api_call**\n**\n: model:\n {model} \n\n messages:\n {messages} \n\n rest: {kwargs}"
        )

        request = kwargs.get("additional_args").get("complete_input_dict").copy()
        request["messages"] = messages.copy()
        completion_id = self.log_start(request, Kind.chat, self.tags)
        last_completion_response_var.set({"completionID": completion_id})
        litellm_call_id = kwargs.get("litellm_call_id")
        self.runs[litellm_call_id] = {
            "kind": Kind.chat,
            "completion_id": completion_id,
            "model": model,
        }

    def log_post_api_call(self, kwargs, response_obj, start_time, end_time):
        pass

    def log_stream_event(self, kwargs, response_obj, start_time, end_time):
        pass

    def log_success_event(self, kwargs, response_obj, start_time, end_time):
        logger.debug(
            f"**\n**log_success_event**\n**\n: response_obj:\n {response_obj} \n\n start_time:\n {start_time} \n\n end_time: {end_time}"
        )
        litellm_call_id = kwargs.get("litellm_call_id")
        run = self.runs.get(litellm_call_id, None)
        duration = (end_time - start_time).total_seconds()

        completion_id = run["completion_id"]
        last_completion_response_var.set({"completionID": completion_id})
        self.log_end(completion_id, response_obj.dict(), duration)

    def log_failure_event(self, kwargs, response_obj, start_time, end_time):
        update_log_row = {
            "status": "failed",
            "failure_kind": type(kwargs["exception"]).__name__,
            "failure_reason": kwargs["exception"].message,
        }

        litellm_call_id = kwargs.get("litellm_call_id")
        run = self.runs.get(litellm_call_id, None)
        self.api_request(
            f"/api/completions/{run['completion_id']}",
            "POST",
            update_log_row,
        )

    #### ASYNC #### - for acompletion/aembeddings

    async def async_log_stream_event(self, kwargs, response_obj, start_time, end_time):
        pass

    async def async_log_success_event(self, kwargs, response_obj, start_time, end_time):
        if kwargs["call_type"] == "completion":
            return

        logger.debug(
            f"**\n**async_log_success_event**\n**\n: kwargs:\n {kwargs} response_obj:\n {response_obj} \n\n {start_time} \n\n duration: {end_time - start_time}"
        )
        self.log_success_event(kwargs, response_obj, start_time, end_time)

    async def async_log_failure_event(self, kwargs, response_obj, start_time, end_time):
        if kwargs["call_type"] == "completion":
            return

        self.log_failure_event(kwargs, response_obj, start_time, end_time)



import json
import re
import string
from copy import deepcopy


def merge_hparams(override, base):
    merged = deepcopy(base)
    if override:
        for hparam in override:
            merged[hparam] = override[hparam]

    return merged


# Ref: https://github.com/openai/evals/blob/a24f20a357ecb3cc5eec8323097aeade9585796c/evals/elsuite/utils.py
def normalize(s: str) -> str:
    """Lower text and remove punctuation, articles and extra whitespace."""
    s = s.lower()
    exclude = set(string.punctuation)
    s = "".join(char for char in s if char not in exclude)
    s = re.sub(r"\b(a|an|the)\b", " ", s)
    s = " ".join(s.split())
    return s


def fuzzy_match(s1: str, s2: str) -> bool:
    s1 = normalize(s1)
    s2 = normalize(s2)

    if s1 == "" or s2 == "":
        return s1 == s2

    return s1 in s2 or s2 in s1


def parse_field(value):
    try:
        # Try to parse the value as JSON (list)
        parsed = json.loads(value)
        if isinstance(parsed, list):
            return parsed
        else:
            return [value]
    except json.JSONDecodeError:
        # If it's not valid JSON, return the original string value as a list with singleton element
        return [value]



import json
import logging

import click
import httpx
from rich.console import Console
from rich.table import Table
from tqdm import tqdm

from log10._httpx_utils import _try_get
from log10.llm import Log10Config


logging.basicConfig(
    format="[%(asctime)s - %(name)s - %(levelname)s] %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
)
logger: logging.Logger = logging.getLogger("LOG10")
logger.setLevel(logging.INFO)


class Feedback:
    feedback_create_url = "/api/v1/feedback"

    def __init__(self, log10_config: Log10Config = None):
        self._log10_config = log10_config or Log10Config()
        self._http_client = httpx.Client()
        self._http_client.headers = {
            "x-log10-token": self._log10_config.token,
            "x-log10-organization-id": self._log10_config.org_id,
            "Content-Type": "application/json",
        }

    def _post_request(self, url: str, json_payload: dict) -> httpx.Response:
        json_payload["organization_id"] = self._log10_config.org_id
        try:
            res = self._http_client.post(self._log10_config.url + url, json=json_payload)
            res.raise_for_status()
            return res
        except Exception as e:
            logger.error(e)
            if hasattr(e, "response") and hasattr(e.response, "json") and "error" in e.response.json():
                logger.error(e.response.json()["error"])
            raise

    def create(
        self,
        task_id: str,
        values: dict,
        completion_tags_selector: list[str],
        comment: str = None,
    ) -> httpx.Response:
        json_payload = {
            "task_id": task_id,
            "json_values": values,
            "completion_tags_selector": completion_tags_selector,
            "comment": comment,
        }
        res = self._post_request(self.feedback_create_url, json_payload)
        return res

    def list(self, offset: int = 0, limit: int = 50, task_id: str = "") -> httpx.Response:
        base_url = self._log10_config.url
        api_url = "/api/v1/feedback"
        url = f"{base_url}{api_url}?organization_id={self._log10_config.org_id}&offset={offset}&limit={limit}&task_id={task_id}"

        # GET feedback
        try:
            res = self._http_client.get(url=url)
            res.raise_for_status()
            return res
        except Exception as e:
            logger.error(e)
            if hasattr(e, "response") and hasattr(e.response, "json") and "error" in e.response.json():
                logger.error(e.response.json()["error"])
            raise

    def get(self, id: str) -> httpx.Response:
        base_url = self._log10_config.url
        api_url = "/api/v1/feedback"
        get_url = f"{base_url}{api_url}/{id}?organization_id={self._log10_config.org_id}"
        res = _try_get(get_url)
        res.raise_for_status()
        if res.status_code != 200:
            raise Exception(f"Error fetching feedback: {res.json()}")
        return res


@click.command()
@click.option("--task_id", prompt="Enter task id", help="Task ID")
@click.option("--values", prompt="Enter task values", help="Feedback in JSON format")
@click.option(
    "--completion_tags_selector",
    prompt="Enter completion tags selector",
    help="Completion tags selector",
)
@click.option("--comment", help="Comment", default="")
def create_feedback(task_id, values, completion_tags_selector, comment):
    """
    Add feedback to a group of completions associated with a task
    """
    click.echo("Creating feedback")
    tags = completion_tags_selector.split(",")
    values = json.loads(values)
    feedback = Feedback().create(task_id=task_id, values=values, completion_tags_selector=tags, comment=comment)
    click.echo(feedback.json())


def _get_feedback_list(offset, limit, task_id):
    total_fetched = 0
    feedback_data = []
    total_feedback = 0
    if limit:
        limit = int(limit)

    try:
        while True:
            fetch_limit = limit - total_fetched if limit else 50
            res = Feedback().list(offset=offset, limit=fetch_limit, task_id=task_id)
            new_data = res.json().get("data", [])
            if total_feedback == 0:
                total_feedback = res.json().get("total", 0)
            if not limit:
                limit = total_feedback
            feedback_data.extend(new_data)

            current_fetched = len(new_data)
            total_fetched += current_fetched
            offset += current_fetched
            if total_fetched >= limit or total_fetched >= total_feedback:
                break
    except Exception as e:
        click.echo(f"Error fetching feedback {e}")
        if hasattr(e, "response") and hasattr(e.response, "json") and "error" in e.response.json():
            click.echo(e.response.json()["error"])
        return []

    return feedback_data


@click.command()
@click.option(
    "--offset", default=0, type=int, help="The starting index from which to begin the feedback fetch. Defaults to 0."
)
@click.option(
    "--limit", default=25, type=int, help="The maximum number of feedback items to retrieve. Defaults to 25."
)
@click.option(
    "--task_id",
    default="",
    type=str,
    help="The specific Task ID to filter feedback. If not provided, feedback for all tasks will be fetched.",
)
def list_feedback(offset, limit, task_id):
    """
    List feedback based on the provided criteria. This command allows fetching feedback for a specific task or across all tasks,
    with control over the starting point and the number of items to retrieve.
    """
    feedback_data = _get_feedback_list(offset, limit, task_id)
    data_for_table = []
    for feedback in feedback_data:
        data_for_table.append(
            {
                "id": feedback["id"],
                "task_name": feedback["task_name"],
                "feedback": json.dumps(feedback["json_values"], ensure_ascii=False),
                "matched_completion_ids": ",".join(feedback["matched_completion_ids"]),
            }
        )
    table = Table(title="Feedback")
    table.add_column("ID")
    table.add_column("Task Name")
    table.add_column("Feedback")
    table.add_column("Completion ID")

    for item in data_for_table:
        table.add_row(item["id"], item["task_name"], item["feedback"], item["matched_completion_ids"])
    console = Console()
    console.print(table)
    console.print(f"Total feedback: {len(feedback_data)}")


@click.command()
@click.option("--id", required=True, help="Get feedback by ID")
def get_feedback(id):
    """
    Get feedback based on provided ID.
    """
    try:
        res = Feedback().get(id)
    except Exception as e:
        click.echo(f"Error fetching feedback {e}")
        if hasattr(e, "response") and hasattr(e.response, "json") and "error" in e.response.json():
            click.echo(e.response.json()["error"])
        return
    console = Console()
    feedback = json.dumps(res.json(), indent=4)
    console.print_json(feedback)


@click.command()
@click.option(
    "--offset",
    default=0,
    help="The starting index from which to begin the feedback fetch. Leave empty to start from the beginning.",
)
@click.option(
    "--limit", default="", help="The maximum number of feedback items to retrieve. Leave empty to retrieve all."
)
@click.option(
    "--task_id",
    default="",
    type=str,
    help="The specific Task ID to filter feedback. If not provided, feedback for all tasks will be fetched.",
)
@click.option(
    "--file",
    "-f",
    type=str,
    required=False,
    help="Path to the file where the feedback will be saved. The feedback data is saved in JSON Lines (jsonl) format. If not specified, feedback will be printed to stdout.",
)
def download_feedback(offset, limit, task_id, file):
    """
    Download feedback based on the provided criteria. This command allows fetching feedback for a specific task or across all tasks,
    with control over the starting point and the number of items to retrieve.
    """
    feedback_data = _get_feedback_list(offset, limit, task_id)

    console = Console()
    if not file:
        for feedback in feedback_data:
            console.print_json(json.dumps(feedback, indent=4))
        return

    with open(file, "w") as f:
        console.print(f"Saving feedback to {file}")
        for feedback in tqdm(feedback_data):
            f.write(json.dumps(feedback) + "\n")



import os

from google.api_core.exceptions import NotFound
from google.cloud import bigquery


# todo: add requirements.txt file
# todo: add instructions for bigquery integration


def initialize_bigquery(debug=False):
    # Configure the BigQuery client
    project_id = os.environ.get("LOG10_BQ_PROJECT_ID")
    dataset_id = os.environ.get("LOG10_BQ_DATASET_ID")
    completions_table_id = os.environ.get("LOG10_BQ_COMPLETIONS_TABLE_ID")

    client = bigquery.Client(project=project_id)

    def dataset_exists(dataset_id):
        try:
            client.get_dataset(dataset_id)  # API request
            return True
        except NotFound:
            return False

    def table_exists(dataset_id, completions_table_id):
        try:
            table_ref = client.dataset(dataset_id).table(completions_table_id)
            client.get_table(table_ref)  # API request
            return True
        except NotFound:
            return False

    # Check if dataset exists
    if dataset_exists(dataset_id):
        if debug:
            print(f"Dataset {dataset_id} exists.")
    else:
        if debug:
            print(f"Dataset {dataset_id} does not exist. Creating...")
        # Create the dataset
        dataset_ref = client.dataset(dataset_id)
        dataset = bigquery.Dataset(dataset_ref)
        # Set the location, e.g., "US", "EU", "asia-northeast1", etc.
        dataset.location = "US"
        created_dataset = client.create_dataset(dataset)  # API request

        if debug:
            print(f"Dataset {created_dataset.dataset_id} created.")

    client_dataset = client.dataset(dataset_id)

    # Check if table exists
    if table_exists(dataset_id, completions_table_id):
        if debug:
            print(f"Table {completions_table_id} exists in dataset {dataset_id}.")
    else:
        if debug:
            print(f"Table {completions_table_id} does not exist in dataset {dataset_id}. Creating...")
        # Create the table
        table_ref = client_dataset.table(completions_table_id)
        script_dir = os.path.dirname(os.path.abspath(__file__))
        schema_path = os.path.join(script_dir, "schemas", "bigquery.json")
        schema = client.schema_from_json(schema_path)
        table = bigquery.Table(table_ref, schema=schema)
        created_table = client.create_table(table)  # API request

        if debug:
            print(f"Table {created_table.table_id} created.")

    # load dataset and table
    table_ref = client_dataset.table(completions_table_id)
    table = client.get_table(table_ref)

    return client, table



import json
import time

import click
import httpx
import pandas as pd
import rich
import tqdm
from rich.console import Console
from rich.table import Table

from log10._httpx_utils import _get_time_diff, _try_get
from log10.cli_utils import generate_markdown_report, generate_results_table
from log10.llm import Log10Config
from log10.prompt_analyzer import PromptAnalyzer, convert_suggestion_to_markdown, display_prompt_analyzer_suggestions


_log10_config = Log10Config()


def _get_completion(id: str) -> httpx.Response:
    url = f"{_log10_config.url}/api/completions/{id}?organization_id={_log10_config.org_id}"
    return _try_get(url)


def _get_tag_id(tag: str) -> str:
    url = f"{_log10_config.url}/api/tags/search?organization_id={_log10_config.org_id}&query={tag}"
    res = _try_get(url)
    data = res.json()["data"]
    if not data:
        return []

    if len(data) == 1:
        return data[0]["id"]
    else:
        for item in data:
            if item["name"] == tag:
                return item["id"]
        return []


def _get_tag_ids(tags):
    tag_ids = []
    for tag in [t for t in tags.split(",") if t]:
        tag_id = _get_tag_id(tag)
        if tag_id:
            tag_ids.append(tag_id)
        else:
            raise SystemExit(f"Cannot found tag: {tag}.")

    tag_ids_str = ",".join(tag_ids)
    return tag_ids_str


def _get_completions_url(limit, offset, tags, from_date, to_date, base_url, org_id, printout=True):
    tag_ids_str = _get_tag_ids(tags) if tags else ""
    if tag_ids_str and printout:
        rich.print(f"Filter with tags: {tags}")

    date_range = _get_valid_date_range(from_date, to_date)
    if date_range and printout:
        rich.print(f"Filter with created date: {date_range['from'][:10]} to {date_range['to'][:10]}")

    url = f"{base_url}/api/completions?organization_id={org_id}&offset={offset}&limit={limit}&tagFilter={tag_ids_str}&createdFilter={json.dumps(date_range)}&sort=created_at&desc=true&ids="
    return url


def _get_valid_date_range(from_date, to_date):
    if not from_date and not to_date:
        return {}

    if from_date is None or to_date is None:
        raise click.UsageError("Both --from and --to must be set together.")

    if from_date >= to_date:
        raise click.UsageError(f"from_date {from_date} must be earlier than to_date {to_date}")

    parsed_from_date = from_date.replace(hour=8).strftime("%Y-%m-%dT%H:%M:%S.%f")[:-3] + "Z"
    parsed_to_date = to_date.replace(hour=8).strftime("%Y-%m-%dT%H:%M:%S.%f")[:-3] + "Z"

    date_range = {"from": parsed_from_date, "to": parsed_to_date}
    return date_range


def _render_completions_table(completions_data, total_completions):
    data_for_table = []
    for completion in completions_data:
        prompt, response = "", ""
        if completion.get("kind") == "completion":
            prompt = completion.get("request", {}).get("prompt", "")
            response_choices = completion.get("response", {}).get("choices", [])
            if response_choices:
                response = response_choices[0].get("text", "")
        elif completion.get("kind") == "chat":
            request_messages = completion.get("request", {}).get("messages", [])
            prompt = request_messages[0].get("content", "") if request_messages else ""

            response_choices = completion.get("response", {}).get("choices", [])
            if response_choices:
                # Handle 'message' and 'function_call' within the first choice safely
                first_choice = response_choices[0]
                if "message" in first_choice:
                    response = first_choice["message"].get("content", "")
                elif "function_call" in first_choice:
                    response = json.dumps(first_choice.get("function_call", {}))
        else:
            rich.print(f"Unknown completion kind: {completion['kind']} for id: {completion['id']}")

        data_for_table.append(
            {
                "id": completion["id"],
                "status": "success" if completion["status"] == "finished" else completion["status"],
                "created_at": _get_time_diff(completion["created_at"]),
                "prompt": prompt,
                "completion": response,
                "tags": [t["name"] for t in completion["tagResolved"]],
            }
        )
    # render data_for_table with rich table
    table = Table(show_header=True, header_style="bold magenta")

    table.add_column("ID", style="dim")
    table.add_column("Status")
    table.add_column("Created At")
    table.add_column("Prompt", overflow="fold")
    table.add_column("Completion", overflow="fold")
    table.add_column("Tags", justify="right")

    max_len = 40
    for item in data_for_table:
        tags = ", ".join(item["tags"]) if item["tags"] else ""
        if isinstance(item["prompt"], list):
            item["prompt"] = " ".join(item["prompt"])
        short_prompt = item["prompt"][:max_len] + "..." if len(item["prompt"]) > max_len else item["prompt"]
        short_completion = (
            item["completion"][:max_len] + "..." if len(item["completion"]) > max_len else item["completion"]
        )
        table.add_row(item["id"], item["status"], item["created_at"], short_prompt, short_completion, tags)

    console = Console()
    console.print(table)
    console.print(f"{total_completions=}")


@click.command()
@click.option("--limit", default=25, help="Specify the maximum number of completions to retrieve.")
@click.option("--offset", default=0, help="Set the starting point (offset) from where to begin fetching completions.")
@click.option(
    "--timeout", default=10, help="Set the maximum time (in seconds) allowed for the HTTP request to complete."
)
@click.option("--tags", default="", help="Filter completions by specific tags. Separate multiple tags with commas.")
@click.option(
    "--from",
    "from_date",
    type=click.DateTime(),
    help="Define the start date for fetching completions (inclusive). Use the format: YYYY-MM-DD.",
)
@click.option(
    "--to",
    "to_date",
    type=click.DateTime(),
    help="Set the end date for fetching completions (inclusive). Use the format: YYYY-MM-DD.",
)
def list_completions(limit, offset, timeout, tags, from_date, to_date):
    """
    List completions
    """
    base_url = _log10_config.url
    org_id = _log10_config.org_id

    url = _get_completions_url(limit, offset, tags, from_date, to_date, base_url, org_id)
    # Fetch completions
    res = _try_get(url, timeout)

    completions = res.json()
    total_completions = completions["total"]
    completions = completions["data"]

    _render_completions_table(completions, total_completions)


@click.command()
@click.option("--id", prompt="Enter completion id", help="Completion ID")
def get_completion(id):
    """
    Get a completion by id
    """
    res = _get_completion(id)
    rich.print_json(json.dumps(res.json()["data"], indent=4))


def _write_completions(res, output_file, compact_mode):
    """Processes completions and appends them to the output file."""
    with open(output_file, "a") as file:
        data = res.json()["data"]
        if compact_mode:
            for completion in data:
                file.write(json.dumps(completion) + "\n")
        else:
            for completion_id in (completion["id"] for completion in data):
                completion = _get_completion(completion_id).json()["data"]
                file.write(json.dumps(completion) + "\n")


@click.command()
@click.option("--limit", default="", help="Specify the maximum number of completions to retrieve.")
@click.option("--offset", default="", help="Set the starting point (offset) from where to begin fetching completions.")
@click.option(
    "--timeout", default=10, help="Set the maximum time (in seconds) allowed for the HTTP request to complete."
)
@click.option("--tags", default="", help="Filter completions by specific tags. Separate multiple tags with commas.")
@click.option(
    "--from",
    "from_date",
    type=click.DateTime(),
    help="Define the start date for fetching completions (inclusive). Use the format: YYYY-MM-DD.",
)
@click.option(
    "--to",
    "to_date",
    type=click.DateTime(),
    help="Set the end date for fetching completions (inclusive). Use the format: YYYY-MM-DD.",
)
@click.option("--compact", is_flag=True, help="Enable to download only the compact version of the output.")
@click.option("--file", "-f", default="completions.jsonl", help="Specify the filename and path for the output file.")
def download_completions(limit, offset, timeout, tags, from_date, to_date, compact, file):
    """
    Download completions to a jsonl file
    """
    base_url = _log10_config.url
    org_id = _log10_config.org_id

    init_url = _get_completions_url(1, 0, tags, from_date, to_date, base_url, org_id)
    res = _try_get(init_url)
    if res.status_code != 200:
        rich.print(f"Error: {res.json()}")
        return

    total_completions = res.json()["total"]
    offset = int(offset) if offset else 0
    limit = int(limit) if limit else total_completions
    rich.print(f"Download total completions: {limit}/{total_completions}")
    if not click.confirm("Do you want to continue?"):
        return

    # dowlnoad completions
    pbar = tqdm.tqdm(total=limit)
    batch_size = 10
    end = offset + limit if offset + limit < total_completions else total_completions
    for batch in range(offset, end, batch_size):
        current_batch_size = batch_size if batch + batch_size < end else end - batch
        download_url = _get_completions_url(
            current_batch_size, batch, tags, from_date, to_date, base_url, org_id, printout=False
        )
        res = _try_get(download_url, timeout)
        _write_completions(res, file, compact)
        pbar.update(current_batch_size)


def _get_llm_repsone(
    model: str,
    messages: list[dict],
    temperature: float = 0.2,
    max_tokens: int = 512,
    top_p: float = 1.0,
):
    ret = {"content": "", "usage": {"prompt_tokens": 0, "completion_tokens": 0, "total_tokens": 0}, "duration": 0.0}

    start_time = time.perf_counter()
    if "gpt-4" in model or "gpt-3.5" in model:
        from log10.load import OpenAI

        client = OpenAI()
        response = client.chat.completions.create(
            model=model, messages=messages, temperature=temperature, max_tokens=max_tokens, top_p=top_p
        )
        ret["content"] = response.choices[0].message.content
        ret["usage"] = response.usage.dict()
    elif "claude-3" in model:
        from log10.load import Anthropic

        system_messages = [m["content"] for m in messages if m["role"] == "system"]
        other_messages = [m for m in messages if m["role"] != "system"]
        system_prompt = ("\n").join(system_messages)

        client = Anthropic()
        response = client.messages.create(
            model=model,
            system=system_prompt,
            messages=other_messages,
            max_tokens=max_tokens,
            temperature=temperature,
            top_p=top_p,
        )
        ret["content"] = response.content[0].text
        ret["usage"]["prompt_tokens"] = response.usage.input_tokens
        ret["usage"]["completion_tokens"] = response.usage.output_tokens
        ret["usage"]["total_tokens"] = response.usage.input_tokens + response.usage.output_tokens
    elif "mistral" in model:
        import mistralai
        from mistralai.client import MistralClient

        from log10.load import log10

        log10(mistralai)

        client = MistralClient()
        response = client.chat(
            model=model, messages=messages, temperature=temperature, max_tokens=max_tokens, top_p=top_p
        )
        ret["content"] = response.choices[0].message.content
        ret["usage"] = response.usage.model_dump()
    else:
        raise ValueError(f"Model {model} not supported.")
    ret["duration"] = int((time.perf_counter() - start_time) * 1000)

    return ret


def _render_comparison_table(model_response_raw_data):
    rich.print(f"completion_id: {model_response_raw_data['completion_id']}")
    rich.print("original_request:")
    rich.print_json(json.dumps(model_response_raw_data["original_request"], indent=4))

    table = rich.table.Table(show_header=True, header_style="bold magenta", box=rich.box.ROUNDED, show_lines=True)
    table.add_column("Model")
    table.add_column("Content")
    table.add_column("Total Token Usage (Input/Output)")
    table.add_column("Duration (ms)")

    for model, data in model_response_raw_data.items():
        # only display model data
        if model not in ["completion_id", "original_request"]:
            usage = data["usage"]
            formatted_usage = f"{usage['total_tokens']} ({usage['prompt_tokens']}/{usage['completion_tokens']})"
            table.add_row(model, data["content"], formatted_usage, str(data["duration"]))
    rich.print(table)


def _create_dataframe_from_comparison_data(model_response_raw_data):
    completion_id = model_response_raw_data["completion_id"]
    original_request = model_response_raw_data["original_request"]
    rows = []
    for model, model_data in model_response_raw_data.items():
        # only display model data
        if model not in ["completion_id", "original_request"]:
            content = model_data["content"]
            usage = model_data["usage"]
            prompt_tokens = usage["prompt_tokens"]
            completion_tokens = usage["completion_tokens"]
            total_tokens = usage["total_tokens"]
            duration = model_data["duration"]
            prompt_messages = json.dumps(original_request["messages"])
            rows.append(
                [
                    completion_id,
                    prompt_messages,
                    model,
                    content,
                    prompt_tokens,
                    completion_tokens,
                    total_tokens,
                    duration,
                ]
            )

    df = pd.DataFrame(
        rows,
        columns=[
            "Completion ID",
            "Prompt Messages",
            "Model",
            "Content",
            "Prompt Tokens",
            "Completion Tokens",
            "Total Tokens",
            "Duration (ms)",
        ],
    )

    return df


def _compare(models: list[str], messages: dict, temperature: float = 0.2, max_tokens: float = 256, top_p: float = 1.0):
    ret = {}
    if models:
        for model in models:
            rich.print(f"Running {model}")
            response = _get_llm_repsone(
                model,
                messages,
                temperature=temperature,
                max_tokens=max_tokens,
                top_p=top_p,
            )

            ret[model] = response
    return ret


_SUPPORTED_MODELS = [
    # openai chat models
    "gpt-4o",
    "gpt-4o-2024-05-13",
    "gpt-4-turbo",
    "gpt-4-turbo-2024-04-09",
    "gpt-4-0125-preview",
    "gpt-4-turbo-preview",
    "gpt-4-1106-preview",
    "gpt-4-vision-preview",
    "gpt-4",
    "gpt-4-0314",
    "gpt-4-0613",
    "gpt-4-32k",
    "gpt-4-32k-0314",
    "gpt-4-32k-0613",
    "gpt-3.5-turbo",
    "gpt-3.5-turbo-16k",
    "gpt-3.5-turbo-0301",
    "gpt-3.5-turbo-0613",
    "gpt-3.5-turbo-1106",
    "gpt-3.5-turbo-0125",
    "gpt-3.5-turbo-16k-0613",
    # anthropic claude
    "claude-3-opus-20240229",
    "claude-3-sonnet-20240229",
    "claude-3-haiku-20240307",
    # mistral
    "mistral-small-latest",
    "mistral-medium-latest",
    "mistral-large-latest",
]


def _check_model_support(model: str) -> bool:
    return model in _SUPPORTED_MODELS


@click.command()
@click.option("--ids", default="", help="Completion IDs. Separate multiple ids with commas.")
@click.option("--tags", default="", help="Filter completions by specific tags. Separate multiple tags with commas.")
@click.option("--limit", help="Specify the maximum number of completions to retrieve filtered by tags.")
@click.option(
    "--offset", help="Set the starting point (offset) from where to begin fetching completions filtered by tags."
)
@click.option("--models", default="", help="Comma separated list of models to compare")
@click.option("--temperature", default=0.2, help="Temperature")
@click.option("--max_tokens", default=512, help="Max tokens")
@click.option("--top_p", default=1.0, help="Top p")
@click.option("--analyze_prompt", is_flag=True, help="Run prompt analyzer on the messages.")
@click.option("--file", "-f", help="Specify the filename for the report in markdown format.")
def benchmark_models(ids, tags, limit, offset, models, temperature, max_tokens, top_p, file, analyze_prompt):
    """
    Compare completions using different models and generate report
    """
    if ids and tags:
        raise click.UsageError("--ids and --tags cannot be set together.")
    if (limit or offset) and not tags:
        raise click.UsageError("--limit and --offset can only be used with --tags.")
    if tags:
        if not limit:
            limit = 5
        if not offset:
            offset = 0

    if not models:
        raise click.UsageError("--models must be set to compare.")
    else:
        for model in [m for m in models.split(",") if m]:
            if not _check_model_support(model):
                raise click.UsageError(f"Model {model} is not supported.")

    # get completions ids
    completion_ids = []
    if ids:
        completion_ids = [id for id in ids.split(",") if id]
    elif tags:
        base_url = _log10_config.url
        org_id = _log10_config.org_id
        url = _get_completions_url(limit, offset, tags, None, None, base_url, org_id)
        res = _try_get(url)
        completions = res.json()["data"]
        completion_ids = [completion["id"] for completion in completions]
        if not completion_ids:
            SystemExit(f"No completions found for tags: {tags}")

    compare_models = [m for m in models.split(",") if m]

    data = []
    skipped_completion_ids = []
    for id in completion_ids:
        # get message from id
        completion_data = _get_completion(id).json()["data"]

        # skip completion if status is not finished or kind is not chat
        if completion_data["status"] != "finished" or completion_data["kind"] != "chat":
            rich.print(f"Skip completion {id}. Status is not finished or kind is not chat.")
            skipped_completion_ids.append(id)
            continue

        original_model_request = completion_data["request"]
        original_model_response = completion_data["response"]
        original_model = original_model_response["model"]
        benchmark_data = {
            "completion_id": id,
            "original_request": original_model_request,
            f"{original_model} (original model)": {
                "content": original_model_response["choices"][0]["message"]["content"],
                "usage": original_model_response["usage"],
                "duration": completion_data["duration"],
            },
        }
        messages = original_model_request["messages"]
        compare_models_data = _compare(compare_models, messages, temperature, max_tokens, top_p)
        benchmark_data.update(compare_models_data)
        data.append(benchmark_data)

    prompt_analysis_data = {}
    if analyze_prompt:
        rich.print("Analyzing prompts")
        for item in data:
            completion_id = item["completion_id"]
            prompt_messages = item["original_request"]["messages"]
            all_messages = "\n\n".join([m["content"] for m in prompt_messages])
            analyzer = PromptAnalyzer()
            suggestions = analyzer.analyze(all_messages)
            prompt_analysis_data[completion_id] = suggestions

    # create an empty dataframe
    all_df = pd.DataFrame(
        columns=[
            "Completion ID",
            "Prompt Messages",
            "Model",
            "Content",
            "Prompt Tokens",
            "Completion Tokens",
            "Total Tokens",
            "Duration (ms)",
        ]
    )

    #
    # Display or save the results
    #
    if not file:
        # display in terminal using rich
        for ret in data:
            _render_comparison_table(ret)
            if analyze_prompt:
                completion_id = ret["completion_id"]
                suggestions = prompt_analysis_data[completion_id]
                rich.print(f"Prompt Analysis for completion_id: {completion_id}")
                display_prompt_analyzer_suggestions(suggestions)
    else:
        # generate markdown report and save to file
        for ret in data:
            df = _create_dataframe_from_comparison_data(ret)
            all_df = pd.concat([all_df, df])
        pivot_df = all_df.pivot(index="Completion ID", columns="Model", values="Content")
        pivot_df["Prompt Messages"] = all_df.groupby("Completion ID")["Prompt Messages"].first()
        # Reorder the columns
        cols = pivot_df.columns.tolist()
        cols = [cols[-1]] + cols[:-1]
        pivot_df = pivot_df[cols]

        pivot_table = generate_results_table(pivot_df, section_name="model comparison")
        all_results_table = generate_results_table(all_df, section_name="All Results")

        prompt_analysis_markdown = ""
        if analyze_prompt:
            prompt_analysis_markdown = "## Prompt Analysis\n\n"
            for completion_id, suggestions in prompt_analysis_data.items():
                prompt_messages = all_df[all_df["Completion ID"] == completion_id]["Prompt Messages"].values[0]
                prompt_analysis_markdown += (
                    f"### Prompt Analysis for completion_id: {completion_id}\n\n{prompt_messages}\n\n"
                )
                prompt_analysis_markdown += convert_suggestion_to_markdown(suggestions)

        # generate the list of skipped completions ids
        skipped_completion_markdown = ""
        if skipped_completion_ids:
            skipped_completion_ids_str = ", ".join(skipped_completion_ids)
            skipped_completion_markdown += "## Skipped Completion IDs\n\n"
            skipped_completion_markdown += f"Skipped completions: {skipped_completion_ids_str}\n\n"

        generate_markdown_report(
            file, [pivot_table, prompt_analysis_markdown, all_results_table, skipped_completion_markdown]
        )
        rich.print(f"Report saved to {file}")
